


#' @title Forward-Backward Selection of the Number of Components \eqn{K}
#' 
#' @description
#' 
#' To compare \eqn{gh}-parsimonious models with different number of components \eqn{K}
#' and select the optimal model using the Vuong's closeness test.
#' 
#' @param object \linkS4class{fmx_QLMDe} object
#' 
#' @param test see parameter \code{test} of \link{Step_fmx} function
#' 
#' @param Kmax \link[base]{integer} scalar \eqn{K_M}, the maximum number of components to be considered
#' 
#' @param ... additional parameters
#' 
#' @details 
#' 
#' \link{StepK_fmx} compares the \eqn{gh}-parsimonious models with different number of components,
#' and selects the optimal model using the Vuong's closeness test.
#' 
#' The forward-backward selection starts with finding the \eqn{gh}-parsimonious model 
#' at a user-specified initial number of components \eqn{K = K_0} (as reflected in the input \code{object}).
#' 
#' The forward selection compares the \eqn{gh}-parsimonious models at \eqn{K_0+1} and at \eqn{K_0} component, respectively,
#' using the Vuong's closeness test. 
#' If \eqn{K_0} component is preferred, then the forward-backward selection is stopped if \eqn{K_0=1},
#' otherwise (if \eqn{K_0>1}) switches to the backward selection.
#' If \eqn{K_0+1} component is preferred, then
#' the algorithm is stopped if \eqn{K_0+1=K_M} (prespecified maximum number of components),
#' otherwise (if \eqn{K_0+1<K_M}) \eqn{K_0+2} versus \eqn{K_0+1} component is compared.
#' 
#' The backward selection is performed only if \eqn{K_0} component 
#' is preferred over \eqn{K_0+1} component. 
#' The \eqn{gh}-parsimonious model at \eqn{K_0-1} and at \eqn{K_0} component, respectively, is compared. 
#' If \eqn{K_0} component is preferred, then the forward-backward selection is stopped.
#' If \eqn{K_0-1} component is preferred, then
#' the forward-backward selection is stopped if \eqn{K_0-1=1},
#' otherwise (if \eqn{K_0-1>1}) \eqn{K_0-2} versus \eqn{K_0} (\strong{not} \eqn{K_0-1}) component is compared.
#' 
#' @return 
#' 
#' \link{StepK_fmx} returns an \linkS4class{fmx_QLMDe} object, with attributes
#' \itemize{
#' \item{\code{anova}} {ANOVA table}
#' \item{\code{objF}} {value of the objective function (either the log-likelihood, AIC or BIC)}
#' }
#' 
#' @examples 
#' (d = fmx('norm', mean = c(1, 4, 8), w = c(3, 3, 4)))
#' x = rfmx(n = 1e3L, dist = d)
#' y1 = QLMDe(x, distname = 'norm', K = 1L)
#' StepK_fmx(y1, Kmax = 3L)
#' 
#' if (FALSE) {
#' # slow, but works
#' (d = fmx('GH', A = c(0, 3), g = c(.2, -.3), h = c(.2, .2), w = c(6, 4)))
#' x = rfmx(n = 1e3L, dist = d)
#' (y1 = QLMDe(x, distname = 'GH', K = 1L))
#' StepK_fmx(y1, Kmax = 3L)
#' }
#' 
#' 
#' @export
StepK_fmx <- function(object, test = c('logLik', 'AIC', 'BIC'), Kmax = stop('must specify maximum `Kmax`'), ...) {
  test <- match.arg(test)
  K <- K_orig <- dim(object@parM)[1L]
  message('Finding parsimonious model at K = ', K)
  modelK <- model_orig <- Step_fmx(object, test = test) # comparison is to parsimonious model at original `K`
  objF <- objF_orig <- attr(model_orig, which = 'objF', exact = TRUE)[1L] # 'list'; objective function for parsimonious model at original `K`
  aod <- aod_orig <- attr(model_orig, which = 'anova', exact = TRUE)[1L, ] # # aod line for parsimonious model at original `K`
  
  compareK <- function(model0, model1, test) {
    # K0 > K1
    K0 <- dim(model0@parM)[1L]
    K1 <- dim(model1@parM)[1L]
    # return `TRUE` indicates selecting `K1` (K_less)
    if (K0 <= K1) stop('`model0` should have higher `K` than `model1`')
    #cat('larger model K0 = ', K0, '\n')
    #cat('smaller model K1 = ', K1, '\n')
    if ((distname <- model0@distname) != model1@distname) stop('`model0` has different `distname` than `model1`')
    objF0 <- attr(model0, which = 'objF', exact = TRUE)[1L]
    objF1 <- attr(model1, which = 'objF', exact = TRUE)[1L]
    switch(test, AIC = , BIC = {
      # smaller the better
      return(unlist(objF1) <= (unlist(objF0) + 1e-07)) # 'TRUE' for selecting `K1` (K_less)
    }, logLik = {
      lr_K <- LikRatio(c(objF0, objF1), # `model0` is more complicated than `model1`
                       type = switch(distname, norm = 'plain', GH = 'vuong', stop('unsupported ', sQuote(object@distname))), 
                       compare = 'first')
      if (inherits(lr_K, what = 'vuong')) {
        # see ?LikRatio: BIC-correction more prone to `model1` (simpler model)
        # 1-comp GH: up to 4; 2-comp normal: 2*2+1=5
        # 2-comp GH: up to 4*2+1=9; 3-comp normal: 3*2+2=8
        return(lr_K[2L, c('Decision_BIC')] != names(objF0))
        # if the more-prone-to-simpler-model-decision suggests `K0` (K_more), then choose K_more
        # otherwise, either '[tie]' (in df, i.e., number of parameters) or `K1` (K_less), then choose K_less
      } else if (inherits(lr_K, what = 'anova')) {
        return(lr_K[2L, length(lr_K)] > .05) # p-value of K_less; 'TRUE' for selecting `K1` (K_less)
      } else stop('should not come here')
      # `lr_K` is not returned, for now
    }, stop('unsupported ', sQuote(test)))
    return(FALSE)
  }
  
  ###################################
  # first, increase `K` (always compare to K-1, not `K_orig`)
  
  while ((K + 1L) <= Kmax) {
    tmp_old <- modelK
    obj_K <- QLMDe(object@data, distname = object@distname, data.name = object@data.name, K = K + 1L, p = object@p)
    message('Finding parsimonious model at K = ', K + 1L)
    tmp_new <- Step_fmx(obj_K, test = test, ...)
    if (compareK(model0 = tmp_new, model1 = tmp_old, test = test)) break # smaller model (i.e. `tmp_old`) is selected 
    K <- K + 1L
    message('Increased K = ', K, ' is selected.')
    modelK <- tmp_new
    aod <- rbind.data.frame(aod, attr(modelK, which = 'anova', exact = TRUE)[1L, ]) # still 'anova'
    objF <- c(objF, attr(modelK, which = 'objF', exact = TRUE)[1L]) # 'list'
  }
  
  ###################################
  # then, decrease `K` (always compare to `K_orig`, not K+1)
  
  if (dim(modelK@parM)[1L] == K_orig) { # only do `descreasing` if no `increasing`
    while (K > 1L) {
      obj_K <- QLMDe(object@data, distname = object@distname, data.name = object@data.name, K = K - 1L, p = object@p)
      message('Finding parsimonious model at K = ', K - 1L)
      tmpK <- Step_fmx(obj_K, test = test, ...)
      if (!compareK(model0 = model_orig, model1 = tmpK, test = test)) break # larger model (i.e. `model_orig` is retained)
      K <- K - 1L
      message('Reduced K = ', K, ' is selected.')
      modelK <- tmpK
      aod <- rbind.data.frame(aod, attr(modelK, which = 'anova', exact = TRUE)[1L, ]) # still 'anova'
      objF <- c(objF, attr(modelK, which = 'objF', exact = TRUE)[1L]) # 'list'
    }
  }
  
  attr(aod, 'heading') <- 'Stepwise #-of-Comp Selection (Parsimonious per #-of-Comp)'
  attr(modelK, 'anova') <- aod
  attr(modelK, 'objF') <- objF
  if (dim(modelK@parM)[1L] != K_orig) attr(modelK, 'orig_K') <- model_orig # parsimonious model at original K
  return(modelK)
  
}




